Lambda Calculus Interpreter
===========================

A lazy graph reduction evaluator for the untyped lambda calculus.

./lci program.lc <input.txt >output.txt


> module Main where

> import qualified Data.Map as M
> import Data.Map(Map)
> import Data.Char(ord, chr)
> import Data.List(nub, (\\))
> import Control.Applicative(
>   Applicative, pure, (<*>),
>   Alternative, empty, (<|>),
>   (<$>), (<$), (<*), some, liftA2)
> import System.Environment(getArgs)


Strip comments
--------------

A comment is everything from '#' to the end of the line.

> decomment :: String -> String
> decomment = concatMap (fst . break ('#'==)) . lines


Tokenize
--------

> digits :: String
> digits = "0123456789"

> letters :: String
> letters = "abcdefghijklmnopqrstuvwxyz"

> symbols :: [Char]
> symbols = "\\.()[],"

Split a string into tokens (each itself a string), such that each token
consists of either all digits, all letters, or a single symbol.  Use
whitespace to separate tokens.

> tokenize :: String -> [String]
> tokenize [] = []
> tokenize (c:cs)
>  | c `elem` digits  = let (t,ts) = span (`elem` digits) cs
>                       in (c:t):tokenize ts
>  | c `elem` letters = let (t,ts) = span (`elem` letters) cs
>                       in (c:t):tokenize ts
>  | c `elem` symbols = [c] : tokenize cs
>  | c == ' ' = tokenize cs
>  | otherwise = []


Parsing primitives
------------------

A parser takes a list of tokens to a list of possible partial parses.

> newtype Parser s t = P{ unP :: [s] -> [(t, [s])] }

> instance Functor (Parser s) where
>   fmap f (P p) = P (\q -> [ (f v, s) | (v,s) <- p q ] )

> instance Applicative (Parser s) where
>   pure f = P (\q -> [(f, q)])
>   P p1 <*> P p2 = P (\q -> [ (u v, t) | (u, s) <- p1 q, (v, t) <- p2 s ])

> instance Alternative (Parser s) where
>   empty = P (\_ -> [])
>   P p1 <|> P p2 = P (\q -> p1 q ++ p2 q)

Accept a token that satisfies a predicate.

> satisfy :: (s -> Bool) -> Parser s s
> satisfy p = P (\q -> case q of
>   (x:xs) | p x -> [(x,xs)]
>   _ -> [])

Accept a specific token.

> sym :: Eq s => s -> Parser s s
> sym a = satisfy (== a)

Accept some p's separated by s's.

> someSep :: Parser s a -> Parser s t -> Parser s [t]
> someSep s p = (:[]) <$> p <|> (:) <$> p <* s <*> someSep s p

Accept some p's separated by s's, or nothing.

> manySep :: Parser s a -> Parser s t -> Parser s [t]
> manySep s p = pure [] <|> someSep s p

Accept a name consisting of letters.

> name :: Parser String String
> name = P (\q -> case q of
>   (p:ps) | all (`elem` letters) p -> [(p, ps)]
>   _ -> [])

Accept an integer consisting of digits.

> integer :: Parser String Integer
> integer = P (\q -> case q of
>   (p:ps) | all (`elem` digits) p -> [(read p, ps)]
>   _ -> [])


Parse sugared terms
-------------------

> type Name = String

Grammar for sugared terms:

  T ::= v | '\\' v+ '.' T | T+ | '(' T ')' | n | '[' (T (, T)*)? ']'

> data STerm = SVariable Name | SLambda [Name] STerm | SApply [STerm]
>            | SGroup STerm | SInteger Integer | SList [STerm]
>   deriving (Show)

> sTerm :: Parser String STerm
> sTerm  =  SApply <$> some sTerm'

> sTerm' :: Parser String STerm
> sTerm' =  SLambda <$ sym "\\" <*> some name <* sym "." <*> sTerm
>       <|> SGroup <$ sym "(" <*> sTerm <* sym ")"
>       <|> SVariable <$> name
>       <|> SInteger <$> integer
>       <|> SList <$ sym "[" <*> manySep (sym ",") sTerm <* sym "]"


Desugar terms
-------------

The unsugared Lambda Calculus:

  T ::= v | (\v.T) | (T T)

The Reference constructor is used later in the lazy evaluator.

> data Term = Variable Name | Lambda Name Term | Apply Term Term
>           | Reference Pointer
>   deriving (Show)

A list of all possible legal variable names, need to generate fresh
variables (ie, names guaranteed to be unused in a given term).

> variables :: [Name]
> variables = [ v:vs | vs <- [] : variables, v <- letters ]

Check if a variable occurs free in a term.

> isFreeIn :: Name -> Term -> Bool
> isFreeIn n (Variable v)  = n == v
> isFreeIn n (Lambda v t)  = if n == v then False else n `isFreeIn` t
> isFreeIn n (Apply t t')  = n `isFreeIn` t || n `isFreeIn` t'
> isFreeIn _ (Reference _) = error $ "isFreeIn: can't handle reference"

Get all variable names defined or referenced by a term.

> variablesIn :: Term -> [Name]
> variablesIn (Variable v)  = [v]
> variablesIn (Lambda v t)  = nub $ v : variablesIn t
> variablesIn (Apply t t')  = nub $ variablesIn t ++ variablesIn t'
> variablesIn (Reference _) = error $ "variablesIn: can't handle reference"

Get all free variables referenced by a term.

> freeVariablesIn :: Term -> [Name]
> freeVariablesIn t = filter (`isFreeIn` t) (variablesIn t)

Fold over natural numbers.

> foldNat :: (a -> a) -> a -> Integer -> a
> foldNat _ z 0 = z
> foldNat s z n = s (foldNat s z (n-1))

Desugar according to the following conventions:

  Outermost parentheses are dropped:
    M N means of (M N)

  Applications are left associative:
    M N P means (M N) P

  The body of an abstraction extends as far right as possible
    \x.M N means \x.(M N) and not (\x.M) N

  A sequence of abstractions are contracted:
    \x.\y.\z.N is abbreviated as \x y z.N

Desugar integers with their Church-encoding:

  N      -> \s z . (s (s ... (s z) ... ) )    (N applications of s)

Desugar lists with their Scott-encoding:

  []     -> \ c n . n
  (x:xs) -> \ c n . c (desugar x) (desugar xs)

> desugar :: STerm -> Maybe Term
> desugar (SVariable v) = Just $ Variable v
> desugar (SLambda [] _) = Nothing
> desugar (SLambda [v] t) = fmap (Lambda v) $ desugar t
> desugar (SLambda (v:vs) t) = fmap (Lambda v) $ desugar (SLambda vs t)
> desugar (SApply []) = Nothing
> desugar (SApply ts) = foldl1 (liftA2 Apply) (map desugar ts)
> desugar (SGroup t) = desugar t
> desugar (SInteger n) = Just $ (Lambda "s" (Lambda "z"
>   (foldNat (Apply (Variable "s")) (Variable "z") n)))
> desugar (SList []) = Just $ Lambda "c" (Lambda "n" (Variable "n"))
> desugar (SList (l:ls)) =
>   let t = desugar l
>       ts = desugar (SList ls)
>   in case (t,ts) of
>     (Just t', Just ts') ->
>       let (c:n:_) = (variables \\ freeVariablesIn t') \\ freeVariablesIn ts'
>       in Just $ Lambda c (Lambda n (Apply (Apply (Variable c) t') ts'))
>     _ -> Nothing


Convert input
-------------

Convert a String to a Scott-encoded list of Church-encoded integers.

> toLambda :: String -> Term
> toLambda []     = Lambda "c" (Lambda "n" (Variable "n"))
> toLambda (x:xs) = Lambda "c" (Lambda "n" (Apply (Apply
>                     (Variable "c") (toLambda' x)) (toLambda xs)))

> toLambda' :: Char -> Term
> toLambda' c = Lambda "s" (Lambda "z"
>                 (foldNat (Apply (Variable "s")) (Variable "z")
>                   (fromIntegral . ord $ c)))


Evaluate by lazy graph reduction
--------------------------------

> type Pointer = Integer
> type Graph = Map Pointer Term

Get an unused pointer.

> nextPointer :: Graph -> Pointer
> nextPointer g = case M.maxViewWithKey g of
>   Nothing        -> 0
>   Just ((k,_),_) -> k + 1

Replace a free variable by a term.

> replace :: Name -> Term -> Term -> Graph -> (Term, Graph)

Replace a variable if it matches.

> replace x v t@(Variable n) g =
>   if x == n
>   then (v, g)
>   else (t, g)

Replace in the body of an abstraction if it doesn't match the variable.

> replace x v t@(Lambda n s) g =
>   if x == n
>   then (t, g)
>   else let (s', g') = replace x v s g in (Lambda n s', g')

Replace in both branches of an application.

> replace x v (Apply a b) g =
>   let (a', g') = replace x v a g
>       (b', g'') = replace x v b g'
>   in (Apply a' b', g'')

Replace inside a reference, updating it.

> replace x v   (Reference p) g =
>   case M.lookup p g of
>     Nothing -> error $ "replace: bad pointer: " ++ show p
>     Just t -> let (t', g') = replace x v t g
>                   g'' = M.insert p t' g'
>               in (Reference p, g'')

Reduce a graph one step, returning Nothing if it is irreducible.

> reduce :: Term -> Graph -> Maybe (Term, Graph)

Variables are irreducible.

> reduce (Variable _) _ = Nothing

Don't reduce inside an abstraction :(

> reduce (Lambda x _) _ = Nothing

Reduce an application.

> reduce (Apply a b) g = case a of

Given (Apply (Lambda ...) ...), then beta reduce.

>   Lambda n a' -> let p  = nextPointer g
>                      g' = M.insert p b g
>                      (t', g'') = replace n (Reference p) a' g'
>                  in Just (t', g'')

Otherwise reduce a inside the application.

>   _           -> case reduce a g of
>     Just (a', g') -> Just (Apply a' b, g')
>     Nothing -> case reduce b g of
>       Just (b', g') -> Just (Apply a b', g')
>       Nothing -> Nothing

Reducing a reference involves reducing its target and updating the graph.

> reduce (Reference p) g = case M.lookup p g of
>   Nothing -> error $ "reduce: bad pointer: " ++ show p
>   Just t -> case reduce t g of
>     Just (t', g') -> let g'' = M.insert p t' g'
>                      in Just (Reference p, g'')
>     Nothing -> Just (t, g)

Reduce a graph until it can't be reduced further.

> normalize :: Term -> Graph -> (Term, Graph)
> normalize t g = case normalize' t g of
>   Nothing -> (t, g)
>   Just (t', g') -> normalize t' g'

If the top level term is a reference, then look it up.

> normalize' :: Term -> Graph -> Maybe (Term, Graph)
> normalize' (Reference p) g = case M.lookup p g of
>   Nothing -> error $ "normalize': bad pointer: " ++ show p
>   Just t -> Just (t, g)
> normalize' t g = reduce t g


Reduce and convert output
-------------------------

Convert a Scott-encoded list of Church-encoded integers to a String.

> fromLambda :: Term -> Graph -> String
> fromLambda i g = case normalize i g of
>   (Lambda c i', g1) -> case normalize i' g1 of
>     (Lambda n t', g2) -> case normalize t' g2 of
>       (Variable v, _) ->
>         if v == n then [] else error "fromLambda: not a list: expected nil"
>       (Apply a ts, g3) -> case normalize a g3 of
>         (Apply a' t, g4) -> case normalize a' g4 of
>           (Variable v, g5) ->
>             if v == c
>             then let (o, g6) = fromLambda' t g5
>                  in o : fromLambda ts g6
>             else error "fromLambda: not a list: expected cons"
>           _ -> error "fromLambda: not a list: expected cons (variable)"
>         _ -> error "fromLambda: not a list: expected cons (apply)"
>       _ -> error "fromLambda: not a list: expected nil or cons"
>     _ -> error $ "fromLambda: not a list (lambda lambda)"
>   _ -> error "fromLambda: not a list (lambda)"

> fromLambda' :: Term -> Graph -> (Char, Graph)
> fromLambda' i g = case normalize i g of
>   (Lambda s i', g') -> case normalize i' g' of
>     (Lambda z t, g'') -> let (n, g''') = fromLambda'' s z t g''
>                          in (chr (fromIntegral n), g''')
>     _ -> error "fromLambda': not an integer"
>   _ -> error "fromLambda': not an integer"

> fromLambda'' :: Name -> Name -> Term -> Graph -> (Integer, Graph)
> fromLambda'' s z i g = case normalize i g of
>   (Variable v, g') ->
>     if v == z
>     then (0, g')
>     else error "fromLambda'': not an integer: expected zero"
>   (Apply a t, g') -> case normalize a g' of
>     (Variable v, g'') ->
>       if v == s
>       then let (n, g''') = fromLambda'' s z t g'' in (1 + n, g''')
>       else error "fromLambda'': not an integer: expected succ"
>     _ -> error "fromLambda'': not an integer: expected succ"
>   _ -> error "fromLambda'': not an integer: expected zero or succ"

> reduceDot :: Term -> Graph -> [String]
> reduceDot t g = toDot t g : case t of
>                               (Lambda _ t') -> reduceDot t' g
>                               _             -> case reduce t g of
>                                                  Nothing       -> []
>                                                  Just (t', g') -> reduceDot t' g'

> toDot :: Term -> Graph -> String
> toDot t g = let (ds, _, _) = toDotT g M.empty (tail variables) (head variables) t
>             in "digraph G { node[shape=\"circle\",color=\"black\",style=\"filled\"]; " ++ concatMap toDotD ds ++ "}\n"

> type NodeID = String
> type Label = String
> data Dot = Node NodeID Label | Edge NodeID NodeID

> toDotT :: Graph -> Map Pointer NodeID -> [NodeID] -> NodeID -> Term -> ([Dot], Map Pointer NodeID, [NodeID])
> toDotT _ ps is       i (Variable _)  = ([Node i "V"], ps, is)
> toDotT g ps (j:is)   i (Lambda _ t)  = let (ds, ps', is') = toDotT g ps is j t
>                                        in ([Node i ("L")] ++ ds ++ [Edge i j], ps', is')
> toDotT g ps (j:k:is) i (Apply a b)   = let (dsa, ps',  is')  = toDotT g ps is  j a
>                                            (dsb, ps'', is'') = toDotT g ps' is' k b
>                                        in ([Node i "A"] ++ dsa ++ dsb ++ [Edge i j, Edge i k], ps'', is'')
> toDotT g ps (j:is)   i (Reference p) = case M.lookup p ps of
>                                          Nothing -> let Just t = M.lookup p g
>                                                         (ds, ps', is') = toDotT g (M.insert p j ps) is j t
>                                                     in ([Node i "R"] ++ ds ++ [Edge i j], ps', is')
>                                          Just k  -> ([Node i "R", Edge i k], ps, (j:is))

> toDotD :: Dot -> String
> toDotD (Node n l) =
>   let c = case l of
>             "V" -> "yellow"
>             "L" -> "cyan"
>             "A" -> "red"
>             "R" -> "blue"
>   in n ++ "[label=\"\",fillcolor=\"" ++ c ++ "\"]; "
> toDotD (Edge m n) = m ++ "->" ++ n ++ "; "


Main program
------------

> main' :: String -> String -> [String]
> main' source input =
>   case unP sTerm . tokenize . decomment $ source of
>     (sterm, []):_ -> case desugar sterm of
>       Just term -> case freeVariablesIn term of
>         [] -> flip (reduceSVG id) M.empty term -- . Apply term . toLambda $ input
>         _ -> error "nonclosed"
>       _ -> error "desugar failed"
>     _ -> error "parse failed"

> main :: IO ()
> main = do
>   args <- getArgs
>   case args of
>     [sourcefile] -> do
>       source <- readFile sourcefile
>       input <- getContents
>       mapM_ (uncurry writeFile) (map ((++".svg") . show) [(1::Integer)..] `zip` main' source input)
>     _ -> error "usage error"

> reduceSVG :: (Term -> Term) -> Term -> Graph -> [SVG]
> reduceSVG top t g = draw (layout (top t) g)
>                   : case t of
>                       Lambda x t' -> reduceSVG (top . Lambda x) t' g
>                       _ ->  case reduce t g of
>                               Just (t', g') -> reduceSVG top t' g'
>                               Nothing       -> []

> data LTerm = LVariable Name Integer Integer
>            | LLambda Name LTerm Integer Integer
>            | LApply LTerm LTerm Integer Integer
>            | LRefInst Pointer LTerm Integer Integer
>            | LReference Pointer Integer Integer
>   deriving (Show)

> coordinates :: LTerm -> (Integer, Integer)
> coordinates (LVariable  _ x y) = (x, y)
> coordinates (LLambda  _ _ x y) = (x, y)
> coordinates (LApply   _ _ x y) = (x, y)
> coordinates (LRefInst _ _ x y) = (x, y)
> coordinates (LReference _ x y) = (x, y)

> layout :: Term -> Graph -> (LTerm, Integer, Integer, Map Pointer (Integer, Integer))
> layout t g = layout' 0 0 M.empty t g

> layout' :: Integer -> Integer -> Map Pointer (Integer, Integer) -> Term -> Graph -> (LTerm, Integer, Integer, Map Pointer (Integer, Integer))
> layout' x y ps (Variable v)  g = (LVariable v x y, 1, 1, ps)
> layout' x y ps (Lambda v t)  g = let (lt, w, h, ps') = layout' x (y + 1) ps t g
>                                      (px, py) = coordinates lt
>                                  in (LLambda v lt px y, w, h + 1, ps')
> layout' x y ps (Apply a b)   g = let (la, aw, ah, psa) = layout' x (y + 1) ps a g
>                                      (lb, bw, bh, psb) = layout' (x + aw + 1) (y + 1) psa b g
>                                  in (LApply la lb (x + aw) y, 1 + aw + bw, 1 + (ah `max` bh), psb)
> layout' x y ps (Reference p) g = if p `M.member` ps
>                                  then (LReference p x y, 1, 1, ps)
>                                  else case M.lookup p g of
>                                         Nothing -> error $ "layout': bad pointer: " ++ show p
>                                         Just t  -> let (lt, w, h, pst) = layout' x (y + 1) ps t g
>                                                        (px, py) = coordinates lt
>                                                    in (LRefInst p lt px y, w, 1 + h, M.insert p (px, py) pst)


> type SVG = String

> circle :: Integer -> Integer -> String -> SVG
> circle x y c = "<circle cx='" ++ show x ++ "' cy='" ++ show y ++ "' r='0.365' fill='" ++ c ++ "' />"

> line :: Integer -> Integer -> Integer -> Integer -> SVG
> line x y x' y' = "<line x1='" ++ show x ++ "' y1='" ++ show y ++ "' x2='" ++ show x' ++ "' y2='" ++ show y' ++ "' />"

> draw :: (LTerm, Integer, Integer, Map Pointer (Integer, Integer)) -> SVG
> draw (t, w, h, ps) =
>   "<?xml version=\"1.0\" standalone=\"no\"?>\n" ++
>   "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n" ++
>   "<svg width='1920px' height='1080px' viewBox='0 0 1920 1080' xmlns='http://www.w3.org/2000/svg' version='1.1'>\n" ++
>   "<rect x='0' y='0' width='1920' height='1080' fill='white' stroke='none' />\n" ++
>   "<g transform='translate(" ++ show dx ++ "," ++ show dy ++ ") scale(" ++ show scale ++ ")'>\n" ++
>   "<g transform='translate(0.5,0.5)'>\n" ++
>   "<g stroke='black' stroke-width='0.1'>\n" ++
>   drawLines ps t ++ drawNodes ps t ++
>   "</g></g></g></svg>\n"
>   where
>     scale, dx, dy :: Double
>     scale = if w * 1080 <= h * 1920 then 1080 / fromIntegral h else 1920 / fromIntegral w
>     dx    = (1920 - scale * fromIntegral w) / 2
>     dy    = (1080 - scale * fromIntegral h) / 2

> drawLines ps (LVariable  _ x y) = ""
> drawLines ps (LLambda  _ t x y) = let (x',y') = coordinates t
>                                   in line x y x' y' ++ drawLines ps t
> drawLines ps (LApply   a b x y) = let (ax, ay) = coordinates a
>                                       (bx, by) = coordinates b
>                                   in line x y ax ay ++ line x y bx by ++ drawLines ps a ++ drawLines ps b
> drawLines ps (LRefInst _ t x y) = let (x',y') = coordinates t
>                                   in line x y x' y' ++ drawLines ps t
> drawLines ps (LReference p x y) = let Just (x',y') = M.lookup p ps
>                                   in line x y x' y'
> drawNodes ps (LVariable  _ x y) = circle x y "yellow"
> drawNodes ps (LLambda  _ t x y) = circle x y "cyan"   ++ drawNodes ps t
> drawNodes ps (LApply   a b x y) = circle x y "red"    ++ drawNodes ps a ++ drawNodes ps b
> drawNodes ps (LRefInst _ t x y) = circle x y "green"  ++ drawNodes ps t
> drawNodes ps (LReference _ x y) = circle x y "blue"
